  /*------------------------------------------------------------------------------------------*\
 *
 *   Copyright (C) 2016 AVM GmbH <fritzbox_info@avm.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
\*------------------------------------------------------------------------------------------*/

#include <linux/linkage.h>
#include <asm/assembler.h>

/* ===   AVM SECURE MONITOR CALL VECTOR  === */

	.arch_extension sec

	.text
	.global avm_secmon_handler_begin
	.global avm_secmon_handler_end

	.balign 32
avm_secmon_handler_begin:
	.long	0x00000000
	.long	0x00000000
ENTRY(avm_secmon_handler)

	/* Secure Config Register holen */
	mrc	p15, 0, r0, c1, c1, 0

	and	r1, r0, #1
	cmp	r1, #1

	addeq	lr, lr, #0xC0000000

	ldr	r2, [lr, #-4]
	and	r2, r2, #1
	cmp	r2, #1

        orreq	r0, r0, #1
	subeq	lr, lr, #0xC0000000
        bicne	r0, r0, #1

	/* Debug
	mrs	r3, cpsr*/

	/* Secure Config Register mit neuem NS-Wert zurueckschreiben */
	mcr	p15, 0, r0, c1, c1, 0

	/* Debug
	str	r3, [r7], #4
	str	r1, [r7], #4
	str	r2, [r7], #4
	str	lr, [r7], #4*/

	/* Rückkehr aus dem SECURE MONITOR */
	movs	pc, lr

ENDPROC(avm_secmon_handler)

	/* Immediate Literale fuer den SECURE MONITOR-Handler */
	.align
/*1:	.long	0xc05bef5c		@ Pointer auf den Stack-Bereich*/
avm_secmon_handler_end:

/* ===   AVM SECURE MONITOR API  === */

ENTRY(avm_secmon_gate)
avm_secmon_gate_begin:

	/* Stack sichern, bei Debug auch noch r9 */
	stmfd	sp!, {r3-r8}

	/* Modus sichern, IRQ/FIQ sperren */
	mrs     r4, cpsr
	cpsid   if

	/* Debug Buffer holen
	ldr	r7, 1f*/

	/* Sichern und Setzen von MVBAR */
	mrc	p15, 0, r5, c12, c0, 1
	ldr	r6, =avm_secmon_handler_begin
	mcr	p15, 0, r6, c12, c0, 1

	/* Adresse sichern */
	mov	r6, r0
	mov	r7, r1
	mov	r8, r2

	/* ================= NON-Secure! ============== */
        /* Monitor call nach avm_secmon_handler, NS = 1 */
	smc #1

        /* Non-Secure Register lesen/schreiben */
	cmp	r6, #1
	ldrne	r6, [r7]
	streq	r8, [r7]
	moveq	r6, r8

        /* Monitor call nach avm_secmon_handler, NS = 0 */
        smc #0
	/* ============================================ */

	/* Rueckgabewert nach R0 */
	mov	r0, r6

	/* Wiederherstellen von MVBAR */
	mcr	p15, 0, r5, c12, c0, 1

	/* Modus wiederherstellen */
	msr	cpsr_fsxc, r4

	/* Stack wiederherstellen, bei Debug auch noch r9 */
	ldmfd	sp!, {r3-r8}

	/* Rückkehr */
	bx	lr

ENDPROC(avm_secmon_gate)

	/* Debug: Immediate Literale fuer den SECURE MONITOR-Handler */
/*1:	.long	0xc05ccc0c		@ Pointer auf den Stack-Bereich*/
avm_secmon_gate_end:

/* === AVM SECURE MONITOR END === */
