#!/bin/bash

function make_links {
    local src_dir=$1
    local dst_dir=$2
    local kbuild_file=$3
    #echo "make_links: $1 | $2 | $3"
    if [ ! -d $dst_dir ]
    then
        mkdir -p $dst_dir
    fi

    for file in $(ls -1 $src_dir/*.h)
    do
        #echo file: $file
        #if [ -f $dst_dir/$(basename "$file") ]; then
        #    echo $dst_dir/$(basename "$file") exists
        #fi
        ln -sf $pwd/$file $dst_dir/$(basename "$file")
        echo "header-y += "$(basename "$file") >> $kbuild_file

    done
}

if [[ $BRCMDRIVERS_DIR == "" ]]
then
    echo "BRCMDRIVERS_DIR not set"
    exit 1
fi

kernel_root_dir=$FRITZ_BOX_BUILD_DIR/$KERNEL_BUILD/kernel/$LINUX_KERNEL_PATH
echo "kernel_root_dir: $kernel_root_dir"
kbuild_file="$kernel_root_dir/include/uapi/Kbuild.bcm"
rm -f $kbuild_file
touch $kbuild_file

RDP_DIR="$FRITZ_BOX_BUILD_DIR/$KERNEL_BUILD/rdp/"

brcm_include_dirs="$PROJECT_DIR/target/bdmf/framework
                   $PROJECT_DIR/target/bdmf/system
                   $PROJECT_DIR/target/bdmf/system/linux
                   $PROJECT_DIR/target/bdmf/system/linux/ce
                   $PROJECT_DIR/target/bdmf/shell
                   $PROJECT_DIR/target/rdpa_gpl
                   $PROJECT_DIR/target/rdp_subsystem
                   $PROJECT_DIR/target/rdp_subsystem/RU
                   $RDP_DIR/drivers/rdpa_gpl/include
                   $RDP_DIR/drivers/bdmf/framework
                   $RDP_DIR/drivers/bdmf/system/linux/ce
                   $RDP_DIR/drivers/rdp_subsystem/rdp
                   $RDP_DIR/drivers/rdp_subsystem/BCM96838
                   $RDP_DIR/drivers/rdp_subsystem/BCM963138
                   $RDP_DIR/drivers/rdpa_gpl/include/autogen
                   $RDP_DIR/drivers/bdmf/system/linux
                   $RDP_DIR/drivers/rdp_subsystem
                   $RDP_DIR/drivers/bdmf/system
                   $RDP_DIR/projects/DSL_63138/drivers/rdd
                   $INC_RDPA_PATH
                   $INC_RDPA_MW_PATH
                   $INC_RDPA_DRV_PATH
                   $INC_BRCMSHARED_PUB_PATH/rdp
                   $INC_BRCMSHARED_PUB_PATH/pmc
                   $INC_BRCMDRIVER_PRIV_PATH/$BRCM_BOARD
                   $BRCMDRIVERS_DIR/broadcom/char/wlcsm_ext/impl1/include
                   $INC_BRCMDRIVER_PUB_PATH/$BRCM_BOARD
                   $INC_BRCMSHARED_PUB_PATH/$BRCM_BOARD
                   $INC_BRCMSHARED_PRIV_PATH/$BRCM_BOARD"

allowed_subdirs="$INC_BRCMDRIVER_PUB_PATH/$BRCM_BOARD/bcm
                 $INC_RDPA_PATH/autogen"
for dir in $brcm_include_dirs
do
	echo "#### DIR: $dir"
    make_links $dir $kernel_root_dir/include $kbuild_file
    for subdir in $(find $dir -type d)
    do
        if [[ $allowed_subdirs =~ .*$subdir.* && $subdir != $dir ]]
        then
            echo "###use subdir ${subdir#$dir}: $subdir"
            echo "header-y += "${subdir#$dir/}/ >> $kbuild_file
			mkdir -p $kernel_root_dir/include/uapi/${subdir#$dir}
            subdir_kbuild=$kernel_root_dir/include/uapi/${subdir#$dir}/Kbuild
            rm -f $subdir_kbuild
            make_links $subdir $kernel_root_dir/include/${subdir#$dir} $subdir_kbuild
        fi
    done
done
exit 0
